from concurrent.futures import ProcessPoolExecutor, as_completed
from glob import glob

import numpy as np
from tqdm import tqdm

from .coordinates import Coordinate
from .database import Database, DataCollection
from .virtual import DataLayout


def open_mfdatabase(paths, engine="netcdf", tolerance=np.timedelta64(0, "us")):
    """
    Open a multiple file database.

    Parameters
    ----------
    paths: str
        The path names given using shell=style wildcards.
    engine: str
        The engine to use to read the file.
    tolerance: timedelta64
        The tolerance to consider that the end of a file is continuous with the begging
        of the following

    ReturnsDA
    -------
    Database
        The database containing all files data.
    """
    fnames = sorted(glob(paths))
    if len(fnames) > 100_000:
        raise NotImplementedError(
            "The maximum number of file that can be opened at once is for now limited "
            "to 100 000."
        )
    with ProcessPoolExecutor() as executor:
        futures = [
            executor.submit(open_database, fname, engine=engine) for fname in fnames
        ]
        dbs = [
            future.result()
            for future in tqdm(
                as_completed(futures),
                total=len(futures),
                desc="Fetching metadata from files",
            )
        ]
    return concatenate(dbs, tolerance=tolerance, virtual=True)


def concatenate(dbs, dim="time", tolerance=None, virtual=False):
    """
    Concatenate several databases along a given dimension.

    Parameters
    ----------
    dbs : list
        List of databases to concatenate.
    dim : str
        The dimension along which concatenate.
    tolerance : float of timedelta64, optional
        The tolerance to consider that the end of a file is continuous with beginning of
        the following, zero by default.
    virtual : bool, optional
        Whether to create a virtual dataset. It requires that all concatenated
        databases are virtual.

    Returns
    -------
    Database
        The concatenated database.
    """
    dbs = sorted(dbs, key=lambda db: db[dim][0])
    axis = dbs[0].get_axis_num(dim)
    dims = dbs[0].dims
    dtype = dbs[0].dtype
    shape = tuple(
        sum([db.shape[a] for db in dbs]) if a == axis else dbs[0].shape[a]
        for a in range(len(dims))
    )
    if virtual:
        data = DataLayout(shape, dtype)
    else:
        data = np.zeros(shape, dtype)
    idx = 0
    tie_indices = []
    tie_values = []
    for db in tqdm(dbs, desc="Linking database"):
        selection = tuple(
            slice(idx, idx + db.shape[axis]) if d == dim else slice(None) for d in dims
        )
        if virtual:
            data[selection] = db.data.vsource
        else:
            data[selection] = db.values
        tie_indices.extend(idx + db[dim].tie_indices)
        tie_values.extend(db[dim].tie_values)
        idx += db.shape[axis]
    coord = Coordinate(tie_indices, tie_values)
    if tolerance is None:
        if np.issubdtype(dbs[0][dim].dtype, np.datetime64):
            tolerance = np.timedelta64(0, "us")
        else:
            tolerance = 0.0
    if not tolerance == 0:
        coord = coord.simplify(tolerance)
    coords = dbs[0].coords
    coords[dim] = coord
    return Database(data, coords)


def open_database(fname, group=None, engine="netcdf", **kwargs):
    """
    Open a database.

    Parameters
    ----------
    fname : str
        The path of the database.
    group : str, optional
        The file group where the database is located, by default None which corresponds
        to the root of the file.
    engine : str, optional
        The file format, by default "netcdf".

    Returns
    -------
    Database
        The opened database.

    Raises
    ------
    ValueError
        If the engine si not recognized.
    """
    if engine == "netcdf":
        return Database.from_netcdf(fname, group=group, **kwargs)
    elif engine == "asn":
        from .io.asn import read

        return read(fname)
    else:
        raise ValueError("engine not recognized")


def open_datacollection(fname, **kwargs):
    """
    Open a DataCollection from a file.

    Parameters
    ----------
    fname : str
        The path of the DataCollection.

    Returns
    -------
    DataCollection
        The opened DataCollection.
    """
    return DataCollection.from_netcdf(fname, **kwargs)



import os
from copy import deepcopy
from tempfile import TemporaryDirectory

import h5py
import numpy as np


class DataSource:
    """
    A lazy array object pointing toward a netCDF4/HDF5 file.
    """

    def __init__(
        self, path_or_dataset, name=None, shape=None, dtype=None, maxshape=None
    ):
        _vsource = h5py.VirtualSource(
            path_or_dataset, name=None, shape=None, dtype=None, maxshape=None
        )
        _slices = tuple([slice(None)] for axis in range(len(_vsource.shape)))
        self._slices = _slices
        self._vsource = _vsource

    def __getitem__(self, key):
        dsource = deepcopy(self)
        if not isinstance(key, tuple):
            key = (key,)
        for k in key:
            if not isinstance(k, slice):
                raise ValueError("only slicing is allowed.")
        for axis, k in enumerate(key):
            dsource._slices[axis].append(k)
        return dsource

    def __array__(self):
        return self.to_layout().__array__()

    def __repr__(self):
        return f"DataSource: {to_human(self.nbytes)} ({self.dtype})"

    @property
    def slices(self):
        return tuple(
            combine_slices(length, *slices)
            for length, slices in zip(self._vsource.shape, self._slices)
        )

    @property
    def vsource(self):
        return self._vsource.__getitem__(self.slices)

    @property
    def shape(self):
        return self.vsource.shape

    @property
    def dtype(self):
        return self.vsource.dtype

    @property
    def path(self):
        return self.vsource.path

    @property
    def name(self):
        return self.vsource.name

    @property
    def id(self):
        return self.vsource.id

    @property
    def sel(self):
        return self.vsource.sel

    @property
    def nbytes(self):
        return np.prod(self.shape) * self.dtype.itemsize

    def to_layout(self):
        layout = DataLayout(self.shape, self.dtype)
        layout[...] = self.vsource
        return layout

    def to_dataset(self, file, name):
        self.to_layout().to_dataset(file, name)


class DataLayout(h5py.VirtualLayout):
    """
    A composite lazy array pointing toward multiple netCDF4/HDF5 files.
    """

    def __array__(self):
        with TemporaryDirectory() as tmpdirname:
            fname = os.path.join(tmpdirname, "vds.h5")
            with h5py.File(fname, "w") as file:
                dataset = file.create_virtual_dataset(
                    "__values__", self, fillvalue=np.nan
                )
            with h5py.File(fname, "r") as file:
                dataset = file["__values__"]
                out = dataset[...]
        return out

    def __repr__(self):
        return f"DataSource: {to_human(self.nbytes)} ({self.dtype})"

    def __getitem__(self, key):
        raise NotImplementedError(
            "Cannot slice DataLayout. Use `self.to_netcdf(fname, virtual=True)` to "
            "write to disk and reopen it with `xdas.open_database(fname)`"
        )

    @property
    def nbytes(self):
        return np.prod(self.shape) * self.dtype.itemsize

    def to_dataset(self, file, name):
        return file.create_virtual_dataset(name, self, fillvalue=np.nan)


def to_human(size):
    unit = {0: "B", 1: "K", 2: "M", 3: "G", 4: "T"}
    n = 0
    while size > 1024:
        size /= 1024
        n += 1
    return f"{size:.1f}{unit[n]}"


def combine_slices(length, *slices):
    r = range(length)
    for s in slices:
        r = r[s]
    if len(r) == 0:
        return slice(0)
    elif r.stop < 0:
        return slice(r.start, None, r.step)
    else:
        return slice(r.start, r.stop, r.step)



import numpy as np
import pytest

from xdas.coordinates import Coordinate, ScaleOffset


class TestScaleOffset:
    def test_init(self):
        transform = ScaleOffset(10.0, 100.0)
        assert transform.scale == 10.0
        assert transform.offset == 100.0

    def test_eq(self):
        transform1 = ScaleOffset(10.0, 100.0)
        transform2 = ScaleOffset(
            np.timedelta64(1, "s"), np.datetime64("2000-01-01T00:00:00")
        )
        assert transform1 == transform1
        assert transform2 == transform2
        assert transform1 != transform2

    def test_direct(self):
        transform = ScaleOffset(10.0, 100.0)
        assert transform.direct(150.0) == 5.0
        assert np.allclose(
            transform.direct(np.array([150.0, 160.0])), np.array([5.0, 6.0])
        )
        transform = ScaleOffset(
            np.timedelta64(1, "s"), np.datetime64("2000-01-01T00:00:00")
        )
        assert transform.direct(np.datetime64("2000-01-01T00:00:10")) == 10.0

    def test_inverse(self):
        transform = ScaleOffset(10.0, 100.0)
        assert transform.inverse(5.0) == 150
        assert np.allclose(
            transform.inverse(np.array([5.0, 6.0])), np.array([150.0, 160.0])
        )
        transform = ScaleOffset(
            np.timedelta64(1, "s"), np.datetime64("2000-01-01T00:00:00")
        )
        assert transform.inverse(10.0) == np.datetime64("2000-01-01T00:00:10")
        assert transform.inverse(10.3) == np.datetime64("2000-01-01T00:00:10")
        assert transform.inverse(10.7) == np.datetime64("2000-01-01T00:00:11")
        assert transform.inverse(11.0) == np.datetime64("2000-01-01T00:00:11")
        transform = ScaleOffset(
            np.timedelta64(1000, "us"), np.datetime64("2000-01-01T00:00:00")
        )
        assert transform.inverse(1000.0) == np.datetime64("2000-01-01T00:00:01")
        assert transform.inverse(1000.3) == np.datetime64("2000-01-01T00:00:01")
        assert transform.inverse(1000.7) == np.datetime64("2000-01-01T00:00:01.001")

    def test_floatize(self):
        assert ScaleOffset.floatize(np.linspace(0, 1)) == ScaleOffset(1.0, 0.0)
        assert ScaleOffset.floatize(
            np.array(
                [
                    np.datetime64("2000-01-01T00:00:00"),
                    np.datetime64("2000-01-02T00:00:00"),
                ]
            )
        ) == ScaleOffset(np.timedelta64(1, "s"), np.datetime64("2000-01-01T12:00:00"))

    def test_accuracy(self):
        x = np.random.rand(1000)
        transform = ScaleOffset.floatize(x)
        assert np.all(np.abs(x - transform.inverse(transform.direct(x))) < 1e-15)

        t0 = np.datetime64("2000-01-01T00:00:00.000000000")
        three_years_ns = t0 + np.random.randint(0, int(1e9) * 3600 * 24 * 1000, 1000)
        three_month_ns = t0 + np.random.randint(0, int(1e9) * 3600 * 24 * 100, 1000)

        t = three_years_ns.astype("datetime64[us]")
        transform = ScaleOffset.floatize(t)
        assert np.all(t == transform.inverse(transform.direct(t)))

        t = three_month_ns
        transform = ScaleOffset.floatize(t)
        assert np.all(t == transform.inverse(transform.direct(t)))

        t = three_years_ns
        with pytest.warns(UserWarning):
            transform = ScaleOffset.floatize(t)
            error = np.abs(t - transform.inverse(transform.direct(t)))
            assert np.all(error < np.timedelta64(1, "us"))


class TestCoordinate:
    def test_init(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert np.allclose(coord.tie_indices, [0, 8])
        assert np.allclose(coord.tie_values, [100.0, 900.0])
        assert coord.kind == "linear"

    def test_bool(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord
        assert not Coordinate([], [])

    def test_len(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert len(coord) == 9
        assert len(Coordinate([], [])) == 0

    def test_repr(self):
        # TODO
        pass

    def test_eq(self):
        coord1 = Coordinate([0, 8], [100.0, 900.0])
        coord2 = Coordinate([0, 8], [100.0, 900.0])
        assert coord1 == coord2

    def test_getitem(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord[0] == 100.0
        assert coord[4] == 500.0
        assert coord[8] == 900.0
        assert coord[-1] == 900.0
        assert coord[-2] == 800.0
        assert np.allclose(coord[[1, 2, 3]], [200.0, 300.0, 400.0])
        with pytest.raises(IndexError):
            coord[9]
            coord[-9]
        assert coord[0:2] == Coordinate([0, 1], [100.0, 200.0])
        assert coord[:] == coord
        assert coord[6:3] == Coordinate([], [])
        assert coord[1:2] == Coordinate([0], [200.0])
        assert coord[-3:-1] == Coordinate([0, 1], [700.0, 800.0])

    def test_setitem(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        with pytest.raises(TypeError):
            coord[1] = 0
            coord[:] = 0

    def test_asarray(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert np.allclose(np.asarray(coord), coord.values)

    def test_dtype(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord.dtype == np.float64

    def test_ndim(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord.ndim == 1
        assert isinstance(coord.ndim, int)

    def test_shape(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord.shape == (9,)

    def test_format_index(self):
        # TODO
        pass

    def test_format_index_slice(self):
        # TODO
        pass

    def test_get_value(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord.get_value(0) == 100.0
        assert coord.get_value(4) == 500.0
        assert coord.get_value(8) == 900.0
        assert coord.get_value(-1) == 900.0
        assert coord.get_value(-9) == 100.0
        assert np.allclose(coord.get_value([1, 2, 3, -2]), [200.0, 300.0, 400.0, 800.0])
        with pytest.raises(IndexError):
            coord.get_value(-10)
            coord.get_value(9)
            coord.get_value(0.5)
        starttime = np.datetime64("2000-01-01T00:00:00")
        endtime = np.datetime64("2000-01-01T00:00:08")
        coord = Coordinate([0, 8], [starttime, endtime])
        assert coord.get_value(0) == starttime
        assert coord.get_value(4) == np.datetime64("2000-01-01T00:00:04")
        assert coord.get_value(8) == endtime
        assert coord.get_value(-1) == endtime
        assert coord.get_value(-9) == starttime

    def test_get_index(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord.get_index(100.0) == 0
        assert coord.get_index(900.0) == 8
        assert coord.get_index(0.0, "nearest") == 0
        assert coord.get_index(1000.0, "nearest") == 8
        assert coord.get_index(125.0, "nearest") == 0
        assert coord.get_index(175.0, "nearest") == 1
        assert coord.get_index(175.0, "before") == 0
        assert coord.get_index(200.0, "before") == 1
        assert coord.get_index(200.0, "after") == 1
        assert coord.get_index(125.0, "after") == 1
        assert np.all(np.equal(coord.get_index([100.0, 900.0]), [0, 8]))
        with pytest.raises(KeyError):
            assert coord.get_index(0.0) == 0
            assert coord.get_index(1000.0) == 8
            assert coord.get_index(150.0) == 0
            assert coord.get_index(1000.0, "after") == 8
            assert coord.get_index(0.0, "before") == 0

        starttime = np.datetime64("2000-01-01T00:00:00")
        endtime = np.datetime64("2000-01-01T00:00:08")
        coord = Coordinate([0, 8], [starttime, endtime])
        assert coord.get_index(starttime) == 0
        assert coord.get_index(endtime) == 8
        assert coord.get_index(str(starttime)) == 0
        assert coord.get_index(str(endtime)) == 8
        assert coord.get_index("2000-01-01T00:00:04.1", "nearest") == 4

    def test_indices(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert np.all(np.equal(coord.indices, np.arange(9)))

    def test_values(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert np.allclose(coord.values, np.arange(100.0, 1000.0, 100.0))

    def test_get_index_slice(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord.get_index_slice(slice(100.0, 200.0)) == slice(0, 2)
        assert coord.get_index_slice(slice(150.0, 250.0)) == slice(1, 2)
        assert coord.get_index_slice(slice(300.0, 500.0)) == slice(2, 5)
        assert coord.get_index_slice(slice(0.0, 500.0)) == slice(0, 5)
        assert coord.get_index_slice(slice(125.0, 175.0)) == slice(1, 1)
        assert coord.get_index_slice(slice(0.0, 50.0)) == slice(0, 0)
        assert coord.get_index_slice(slice(1000.0, 1100.0)) == slice(9, 9)
        assert coord.get_index_slice(slice(1000.0, 500.0)) == slice(9, 5)
        assert coord.get_index_slice(slice(None, None)) == slice(None, None)

    def test_slice_index(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        assert coord.slice_index(slice(0, 2)) == Coordinate([0, 1], [100.0, 200.0])
        assert coord.slice_index(slice(7, None)) == Coordinate([0, 1], [800.0, 900.0])
        assert coord.slice_index(slice(None, None)) == coord
        assert coord.slice_index(slice(0, 0)) == Coordinate([], [])
        assert coord.slice_index(slice(4, 2)) == Coordinate([], [])
        assert coord.slice_index(slice(9, 9)) == Coordinate([], [])
        assert coord.slice_index(slice(3, 3)) == Coordinate([], [])
        assert coord.slice_index(slice(0, -1)) == Coordinate([0, 7], [100.0, 800.0])
        assert coord.slice_index(slice(0, -2)) == Coordinate([0, 6], [100.0, 700.0])
        assert coord.slice_index(slice(-2, None)) == Coordinate([0, 1], [800.0, 900.0])
        assert coord.slice_index(slice(1, 2)) == Coordinate([0], [200.0])
        assert coord.slice_index(slice(1, 3, 2)) == Coordinate([0], [200.0])
        assert coord.slice_index(slice(None, None, 2)) == Coordinate(
            [0, 4], [100.0, 900.0]
        )
        assert coord.slice_index(slice(None, None, 3)) == Coordinate(
            [0, 2], [100.0, 700.0]
        )
        assert coord.slice_index(slice(None, None, 4)) == Coordinate(
            [0, 2], [100.0, 900.0]
        )
        assert coord.slice_index(slice(None, None, 5)) == Coordinate(
            [0, 1], [100.0, 600.0]
        )
        assert coord.slice_index(slice(2, 7, 3)) == Coordinate([0, 1], [300.0, 600.0])

    def test_to_index(self):
        # TODO
        pass

    def test_simplify(self):
        # TODO
        pass



import dask.array as da
import h5py
import numpy as np
import scipy.signal as sp
import xarray as xr

from xdas.core import Coordinate, Coordinates, Database


def read(fname, decimation=None, preprocess=None):
    """Read a febus A1-R file into a InterpolatedDataArray"""
    data, time, delta, name = read_hdf5(fname)
    if preprocess is not None:
        data, time, delta, name = preprocess(data, time, delta, name)
    if decimation is not None:
        data, delta = decimate(data, delta, decimation)
    data, time = trim_overlaps(data, time, delta)
    time_coord = get_time_coord(data, time, delta)
    offset_coord = get_offset_coord(data, delta)
    data = da.reshape(data, (-1, data.shape[2]))
    database = pack(data, time_coord, offset_coord, name)
    return database


def read_hdf5(fname):
    file = h5py.File(fname, "r")
    (device_name,) = list(file.keys())
    device = file[device_name]
    source = device["Source1"]
    time = np.asarray(source["time"])
    zone = source["Zone1"]
    (name,) = list(zone.keys())
    delta = {
        "time": zone.attrs["Spacing"][1] / 1000.0,
        "offset": zone.attrs["Spacing"][0],
    }
    data = da.from_array(zone[name])
    name = to_snakecase(name)
    return data, time, delta, name


def trim_overlaps(data, time, delta):
    noverlap = data.shape[1] // 4
    data = data[:, noverlap:-noverlap, :]
    time = time + noverlap * delta["time"]
    return data, time


def get_time_coord(data, time, delta):
    starts = np.rint(1e6 * time).astype("datetime64[us]")
    size = data.shape[1]
    step = np.rint(1e6 * delta["time"]).astype("timedelta64[us]")
    return to_coordinate(starts, size, step)


def get_offset_coord(data, delta):
    starts = 0.0
    size = data.shape[2]
    step = delta["offset"]
    return to_coordinate(starts, size, step)


def decimate(data, delta, factor):
    if data.shape[1] % factor != 0:
        raise ValueError(
            "the length of the blocks must be a multitple of the decimation " "factor"
        )
    if 20 * factor > data.shape[1] // 2:
        raise ValueError(
            "the length of the FIR coeffs must be at most half the length of "
            "the blocks"
        )
    data = da.blockwise(
        lambda x: sp.decimate(x, factor, ftype="fir", axis=1),
        "ijk",
        data,
        "ijk",
        adjust_chunks={"j": lambda n: n // factor},
        dtype=data.dtype,
    )
    delta["time"] = factor * delta["time"]
    return data, delta


def pack(data, time_coord, offset_coord, name):
    dims = ("time", "offset")
    coords = Coordinates({"time": time_coord, "offset": offset_coord})
    coords["time"].simplify(np.timedelta64(0, "us"))
    data_array = xr.DataArray(data, dims=dims, name=name)
    database = Database(data_array, coords)
    return database


def to_snakecase(name):
    name = "".join(["_" + c.lower() if c.isupper() else c for c in name]).lstrip("_")
    return name


def to_coordinate(starts, sizes, steps):
    starts, sizes, steps = np.broadcast_arrays(starts, sizes, steps)
    start_values = starts
    end_values = starts + steps * (sizes - 1)
    start_indices = np.insert(np.cumsum(sizes)[:-1], 0, 0)
    end_indices = np.cumsum(sizes) - 1
    tie_values = np.stack((start_values, end_values)).T.reshape(-1)
    tie_indices = np.stack((start_indices, end_indices)).T.reshape(-1)
    time_coordinate = Coordinate(tie_indices, tie_values)
    return time_coordinate


def correct_gps_time(data, time, delta, name):
    dt = data.shape[1] * delta["time"] / 2
    n = data.shape[0]
    t0 = time[0] + np.median(time - time[0] - dt * np.arange(n))
    time = t0 + dt * np.arange(n)
    return data, time, delta, name



import copy
import os
import re

import h5py
import numpy as np
import xarray as xr

from .coordinates import Coordinate, Coordinates
from .virtual import DataLayout, DataSource


class DataCollection(dict):
    """
    A collection of databases.

    A data collection is a dictionary whose keys are any user defined identifiers and
    values are database objects.
    """

    def to_netcdf(self, fname, virtual=False):
        if os.path.exists(fname):
            os.remove(fname)
        for key in self:
            self[key].to_netcdf(fname, group=key, virtual=virtual, mode="a")

    @classmethod
    def from_netcdf(cls, fname):
        with h5py.File(fname, "r") as file:
            groups = list(file.keys())
        self = cls()
        for group in groups:
            self[group] = Database.from_netcdf(fname, group=group)
        return self


class Database:
    """
    N-dimensional array with labeled coordinates and dimensions.

    It is the equivalent of and xarray.DataArray but with custom coordinate objects.
    Most of the Database API follows the DataArray one. Database objects also provide
    virtual dataset capabilities to manipulate huge multi-file NETCDF4 or HDF5 datasets.

    Parameters
    ----------
    data : array_like
        Values of the array. Can be a DataSource or a DataLayout for lazy loading of
        netCDF4/HSF5 files.
    coords : dict of Coordinate
        Coordinates to use for indexing along each dimension.
    dims : sequence of string, optional
        Name(s) of the data dimension(s). If provided, must be equal to the keys of
        `coords`. Used for API compatibility with xarray.
    name : str, optional
        Name of this array.
    attrs : dict_like, optional
        Attributes to assign to the new instance.

    Raises
    ------
    ValueError
        If dims do not match the keys of coords.
    """

    def __init__(self, data, coords, dims=None, name=None, attrs=None):
        # if not (data.shape == tuple(len(coord) for coord in coords.values())):
        # raise ValueError("Shape mismatch between data and coordinates")
        self.data = data
        self.coords = Coordinates(coords)
        self.dims = tuple(coords.keys())
        if dims is not None:
            if not (self.dims == dims):
                raise ValueError("Dimension mismatch between coordinates and dims")
        self.name = name
        self.attrs = attrs

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.coords[key]
        else:
            query = self.coords.get_query(key)
            data = self.data.__getitem__(tuple(query.values()))
            dct = {dim: self.coords[dim][query[dim]] for dim in query}
            coords = Coordinates(dct)
            return self.__class__(data, coords)

    def __setitem__(self, key, value):
        if isinstance(key, str):
            self.coords[key] = value
        else:
            query = self.coords.get_query(key)
            self.data.__setitem__(tuple(query.values()), value)

    def __repr__(self):
        string = "<xdas.Database ("
        string += ", ".join([f"{dim}: {size}" for dim, size in self.sizes.items()])
        string += ")>\n"
        string += repr(self.data) + "\n" + repr(self.coords)
        return string

    def __array__(self):
        return self.data.__array__()

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        raise NotImplementedError()

    def __array_function__(self, func, types, args, kwargs):
        raise NotImplementedError()

    @property
    def shape(self):
        return self.data.shape

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def ndim(self):
        return len(self.dims)

    @property
    def sizes(self):
        return {dim: len(coord) for dim, coord in self.coords.items()}

    @property
    def values(self):
        return self.__array__()

    @property
    def loc(self):
        return LocIndexer(self)

    def get_axis_num(self, dim):
        """
        Return axis number corresponding to dimension in this array.

        Parameters
        ----------
        dim : str
            Dimension name for which to lookup axis.

        Returns
        -------
        int
            Axis number corresponding to the given dimension
        """
        return self.dims.index(dim)

    def isel(self, indexers=None, **indexers_kwargs):
        """
        Return a new Database whose data is given by selecting indexes along the
        specified dimension(s).

        Parameters
        ----------
        indexers : dict, optional
            A dict with keys matching dimensions and values given by integers, slice
            objects or arrays.
        **indexers_kwargs : dict, optional
            The keyword arguments form of integers. Overwrite indexers input if both
            are provided.

        Returns
        -------
        Database
            The selected subset of the Database.
        """
        if indexers is None:
            indexers = {}
        indexers.update(indexers_kwargs)
        return self[indexers]

    def sel(self, indexers=None, **indexers_kwargs):
        """
        Return a new Database whose data is given by selecting index labels along the
        specified dimension(s).

        In contrast to Database.isel, indexers for this method should use labels
        instead of integers.

        Parameters
        ----------
        indexers : dict, optional
            A dict with keys matching dimensions and values given by scalars, slices or
            arrays of tick labels.
        **indexers_kwargs : dict, optional
            The keyword arguments form of integers. Overwrite indexers input if both
            are provided.

        Returns
        -------
        Database
            _description_
        """
        if indexers is None:
            indexers = {}
        indexers.update(indexers_kwargs)
        return self.loc[indexers]

    def copy(self, deep=True, data=None):
        """
        Returns a copy of this array

        If deep=True, a deep copy is made of the data array. Otherwise, a shallow copy
        is made, and the returned data array's values are a new view of this data
        array's values.

        Use data to create a new object with the same structure as original but
        entirely new data.

        Parameters
        ----------
        deep : bool, optional
            Whether the data array and its coordinates are loaded into memory and copied
            onto the new object. Default is True.
        data : array_like, optional
            Data to use in the new object. Must have same shape as original. When data
            is used, deep is ignored for all data variables, and only used for coords.

        Returns
        -------
        DataArray
            New object with dimensions, attributes, coordinates, name, encoding, and
            optionally data copied from original.
        """
        if deep:
            copy_fn = copy.deepcopy
        else:
            copy_fn = copy.copy
        if data is None:
            data = copy_fn(self.data)
        return self.__class__(
            data,
            copy_fn(self.coords),
            copy_fn(self.dims),
            copy_fn(self.name),
            copy_fn(self.attrs),
        )

    def to_xarray(self):
        """
        Convert the Database to a DataArray object.

        Coordinates are converted to dense arrays and lazy values are loaded in memory.

        Returns
        -------
        DataArray
            The converted in-memory DataArray.
        """
        return xr.DataArray(
            data=self.__array__(),
            coords={dim: self.coords[dim].__array__() for dim in self.coords},
            dims=self.dims,
            name=self.name,
            attrs=self.attrs,
        )

    def to_netcdf(self, fname, group=None, virtual=False, **kwargs):
        """
        Write Database contents to a netCDF file.

        Parameters
        ----------
        fname : str
            Path to which to save this dataset.
        group : str, optional
            Path to the netCDF4 group in the given file to open.
        virtual : bool, optional
            Weather to write a virtual dataset. The Database data must be a DataSource
            or a DataLayout. Default is False.

        Raises
        ------
        ValueError
            _description_
        """
        data_vars = []
        mapping = ""
        for dim in self.coords.dims:
            mapping += f"{dim}: {dim}_indices {dim}_values "
            interpolation = xr.DataArray(
                name=f"{dim}_interpolation",
                attrs={
                    "interpolation_name": self.coords[dim].kind,
                    "tie_points_mapping": f"{dim}_points: {dim}_indices {dim}_values",
                },
            ).astype("i4")
            indices = xr.DataArray(
                name=f"{dim}_indices",
                data=self.coords[dim].tie_indices,
                dims=f"{dim}_points",
            )
            values = xr.DataArray(
                name=f"{dim}_values",
                data=self.coords[dim].tie_values,
                dims=f"{dim}_points",
            )
            data_vars.extend([interpolation, indices, values])
        dataset = xr.Dataset(
            data_vars={xarr.name: xarr for xarr in data_vars},
            attrs={"Conventions": "CF-1.9"},
        )
        if not virtual:
            xarr = xr.DataArray(
                self.values,
                dims=self.dims,
                name=self.name,
                attrs={"coordinate_interpolation": mapping},
            )
            dataset[xarr.name] = xarr
            dataset.to_netcdf(fname, group=group, **kwargs)
        elif virtual and isinstance(self.data, (DataSource, DataLayout)):
            xarr = xr.DataArray(
                data=np.empty((0, 0)),
                dims=self.dims,
                name="__tmp__",
                attrs={"coordinate_interpolation": mapping},
            )
            dataset[xarr.name] = xarr
            dataset.to_netcdf(fname, group=group, **kwargs)
            with h5py.File(fname, "r+") as file:
                if self.name is None:
                    name = "__values__"
                else:
                    name = self.name
                if group:
                    file = file[group]
                self.data.to_dataset(file, name)
                for axis, dim in enumerate(self.dims):
                    file[name].dims[axis].attach_scale(file[dim])
                for key in file["__tmp__"].attrs:
                    file[name].attrs[key] = file["__tmp__"].attrs[key]
                del file["__tmp__"]
        else:
            raise ValueError(
                "can only use `virtual=True` with a DataSource or a DataLayout"
            )

    @classmethod
    def from_netcdf(cls, fname, group=None, **kwargs):
        with xr.open_dataset(fname, group=group, **kwargs) as dataset:
            data_vars = [
                var
                for var in dataset.values()
                if "coordinate_interpolation" in var.attrs
            ]
            if len(data_vars) == 1:
                da = data_vars[0]
            else:
                raise ValueError("several possible data arrays detected")
            name = da.name
            coords = Coordinates()
            mapping = da.attrs.pop("coordinate_interpolation")
            matches = re.findall(r"(\w+): (\w+) (\w+)", mapping)
            for match in matches:
                dim, indices, values = match
                coords[dim] = Coordinate(dataset[indices], dataset[values])
        with h5py.File(fname) as file:
            if group:
                file = file[group]
            if name is None:
                name = "__values__"
            data = DataSource(file[name])
        return cls(data, coords)


class LocIndexer:
    def __init__(self, obj):
        self.obj = obj

    def __getitem__(self, key):
        key = self.obj.coords.to_index(key)
        return self.obj[key]

    def __setitem__(self, key, value):
        key = self.obj.coords.to_index(key)
        self.obj[key] = value



import h5py
import numpy as np

from ..coordinates import Coordinate
from ..database import Database
from ..virtual import DataSource


def read(fname):
    with h5py.File(fname, "r") as file:
        header = file["header"]
        t0 = np.datetime64(round(header["time"][()] * 1e6), "us")
        dt = np.timedelta64(round(1e6 * header["dt"][()]), "us")
        dx = header["dx"][()] * np.median(np.diff(header["channels"]))
        data = DataSource(file["data"])
    nt, nd = data.shape
    time = Coordinate([0, nt - 1], [t0, t0 + (nt - 1) * dt])
    distance = Coordinate([0, nd - 1], [0.0, (nd - 1) * dx])
    return Database(data, {"time": time, "distance": distance})



from concurrent.futures import FIRST_COMPLETED, ThreadPoolExecutor, wait

import numpy as np
import scipy.signal as sp

from xdas.core import concatenate


class SignalProcessingChain:
    """
    Chain of processing unit to apply to a Database.

    Parameters
    ----------
    filters : list of callable
        The chain of filters to apply. The filters can have some state memory to ensure
        continuity when repeatedly applied to consecutive chunks.
    """

    def __init__(self, filters):
        self.filters = filters

    def __call__(self, xarr):
        out = xarr
        for filter in self.filters:
            out = filter(out)
        return out

    def process(self, db, dim, chunk_size, parallel=True):
        """
        Process a Database by chunk along a given dimension.

        Parameters
        ----------
        db : Database
            The Database to process.
        dim : str
            Name of the the dimension along which to process the Database.
        chunk_size : int
            The number of samples per chunk.
        parallel : bool, optional
            Whether to use multiprocessing to improve performance, by default True.

        Returns
        -------
        Database
            The processed Database.
        """
        div, mod = divmod(db.sizes[dim], chunk_size)
        if mod == 0:
            nchunks = div
        else:
            nchunks = div + 1
        result = []
        if parallel:
            chunks = []
            for k in range(nchunks):
                query = {dim: slice(k * chunk_size, (k + 1) * chunk_size)}
                chunks.append(db[query])
            scheduler = Scheduler(chunks, self.filters)
            result = scheduler.compute()
        else:
            for k in range(nchunks):
                query = {dim: slice(k * chunk_size, (k + 1) * chunk_size)}
                result.append(self(db[query]))
        return concatenate(result, dim)


class Scheduler:
    """
    Sequential multithreading scheduler.

    As soon as one filter finishes to process a chunk, the scheduler submit the two
    newly available task: applying the same filter on the next chunk and applying the
    next filter to the same chunk.

    Parameters
    ----------
    chunks : list of Database
        The chunks to process.
    filters : list of callable.
        The chain of filters to apply. The filters can have some state memory to ensure
        continuity when repeatedly applied to consecutive chunks.
    """

    def __init__(self, chunks, filters):
        self.chunks = chunks
        self.filters = filters
        self.buffer = {idx: chunk for idx, chunk in enumerate(chunks)}
        self.executor = None
        self.futures = {}

    def compute(self):
        """
        Launch the processing.

        Returns
        -------
        list of Database
            The processed chunks.
        """
        with ThreadPoolExecutor() as executor:
            self.executor = executor
            task = {"chunk": 0, "filter": 0}
            future = self.submit_task(task)
            while self.futures:
                done, _ = wait(self.futures, return_when=FIRST_COMPLETED)
                for future in done:
                    task = self.futures.pop(future)
                    self.buffer[task["chunk"]] = future.result()
                    next_tasks = self.get_next_tasks(task)
                    for next_task in next_tasks:
                        self.submit_task(next_task)
                print("Number of futures:", len(self.futures))
        return list(self.buffer.values())

    def submit_task(self, task):
        """
        Submit a task.

        Parameters
        ----------
        task : dict
            The task to submit described by the number of the filter and the number of
            the chunk.

        Returns
        -------
        future
            The future related to the submitted task.
        """
        filter = self.filters[task["filter"]]
        chunk = self.buffer[task["chunk"]]
        future = self.executor.submit(filter, chunk)
        self.futures[future] = task
        print("New task:", task)
        return future

    def get_next_tasks(self, task):
        """
        Given a finished task, gives the two new available possible tasks: applying the
        same filter on the next chunk and applying the next filter to the same chunk.

        Parameters
        ----------
        task : dict
            The finished task described as the number of the filter and the number of
            the chunk.

        Returns
        -------
        list of dict
            The new available tasks.
        """
        next_tasks = []
        if not task["chunk"] >= len(self.chunks) - 1:
            next_task = task.copy()
            next_task["chunk"] += 1
            next_tasks.append(next_task)
        if not task["filter"] >= len(self.filters) - 1:
            next_task = task.copy()
            next_task["filter"] += 1
            next_tasks.append(next_task)
        return next_tasks


class SignalProcessingUnit:
    """
    Parent class to all signal processing units.
    """

    pass


class LFilter(SignalProcessingUnit):
    """
    Filter data along one-dimension with an IIR or FIR.

    Once the filter is initialized it can be used repeatedly on subsequent chunks of
    data as a normal python function. The state of the filter will be saved between
    each call to ensure continuity of the filtering. The state of the filter can be
    reset for a fresh start.

    Parameters
    ----------
    b : array_like
        The numerator coefficient vector in a 1-D sequence.
    a : array_like
        The denominator coefficient vector in a 1-D sequence.
    dim : str
        Name of the dimension of the input database along which to apply the linear
        filter.
    """

    def __init__(self, b, a, dim):
        self.b = b
        self.a = a
        self.dim = dim
        self.zi = None
        self.axis = None

    def __call__(self, db):
        if (self.zi is None) or (self.axis is None):
            self.initialize(db)
        data, self.zi = sp.lfilter(self.b, self.a, db.data, axis=self.axis, zi=self.zi)
        return db.copy(data=data)

    def initialize(self, db):
        """
        Initialize the filter state.

        Parameters
        ----------
        db : Database
            Some sample of the kind of database to process to get the correct axis
            number.
        """
        self.axis = db.get_axis_num(self.dim)
        zi = sp.lfilter_zi(self.b, self.a)
        n_sections = zi.shape[0]
        s = [n_sections if k == self.axis else 1 for k in range(db.ndim)]
        zi = zi.reshape(s)
        s = [n_sections if k == self.axis else db.shape[k] for k in range(db.ndim)]
        zi = np.broadcast_to(zi, s)
        self.zi = zi

    def reset(self):
        """
        Reset state of the filter for a fresh start.
        """
        self.zi = None
        self.axis = None


class SOSFilter:
    """
    Filter data along one dimension using cascaded second-order sections.

    Parameters
    ----------
    sos : array_like
        Array of second-order filter coefficients.
    dim : str
        Name of the dimension of the input database along which to apply the linear
        filter.
    """

    def __init__(self, sos, dim):
        self.sos = sos
        self.dim = dim
        self.zi = None
        self.axis = None

    def __call__(self, xarr):
        if (self.zi is None) or (self.axis is None):
            self.initialize(xarr)
        data, self.zi = sp.sosfilt(self.sos, xarr, axis=self.axis, zi=self.zi)
        return xarr.copy(data=data)

    def initialize(self, xarr):
        """
        Initialize the filter state.

        Parameters
        ----------
        db : Database
            Some sample of the kind of database to process to get the correct axis
            number.
        """
        self.axis = xarr.get_axis_num(self.dim)
        zi = sp.sosfilt_zi(self.sos)
        ndim = len(xarr.shape)
        n_sections = zi.shape[0]
        s = [n_sections] + [2 if k == self.axis else 1 for k in range(ndim)]
        zi = zi.reshape(s)
        s = [n_sections] + [2 if k == self.axis else xarr.shape[k] for k in range(ndim)]
        zi = np.broadcast_to(zi, s)
        self.zi = zi

    def reset(self):
        """
        Reset state of the filter for a fresh start.
        """
        self.zi = None
        self.axis = None


class Decimate(SignalProcessingUnit):
    def __init__(self, q, dim):
        self.q = q
        self.dim = dim

    def __call__(self, db):
        return db[{self.dim: slice(None, None, self.q)}]


class Writter(SignalProcessingUnit):
    def __init__(self, fname, duration=np.timedelta64(1, "m")):
        self.fname = fname
        self.duration = duration
        self.buffer = None

    def __call__(self, xarr):
        if self.buffer is None:
            self.buffer = xarr
        else:
            self.buffer = concatenate([self.buffer, xarr])
        if self.buffer["time"][-1] - self.buffer["time"][-1] > self.duration:
            self.buffer.sel(
                time=slice(
                    None,
                )
            )



import numpy as np

from xdas.coordinates import Coordinate, Coordinates
from xdas.database import Database


class TestDatabase:
    def generate(self):
        coord = Coordinate([0, 8], [100.0, 900.0])
        coords = Coordinates(dim=coord)
        data = 0.1 * np.arange(9)
        database = Database(data, coords)
        return database

    def test_init_and_properties(self):
        database = self.generate()
        assert database.dims == ("dim",)
        assert database.ndim == 1
        assert database.shape == (9,)
        assert database.sizes == {"dim": 9}
        assert np.allclose(database.data, 0.1 * np.arange(9))
        assert np.all(np.equal(database.values, database.data))
        assert database.get_axis_num("dim") == 0
        assert database.dtype == np.float64

    def test_getitem(self):
        database = self.generate()
        assert database[0].data == 0.0
        assert database[1].data == 0.1
        assert database[0]["dim"] == 100.0
        subdatabase = database[2:4]
        assert np.allclose(subdatabase.data, [0.2, 0.3])
        assert np.allclose(subdatabase["dim"].tie_indices, [0, 1])
        assert np.allclose(subdatabase["dim"].tie_values, [300.0, 400.0])

    def test_setitem(self):
        database = self.generate()
        database[0] = -100.0
        assert database[0].data == -100.0

    def test_sel(self):
        database = self.generate()
        sub = database.sel(dim=slice(2, 4))

    def test_to_xarray(self):
        database = self.generate()
        database.to_xarray()



import warnings

import numpy as np


class Coordinates(dict):
    """
    A dictionary whose keys are dimension names and values are Coordinate objects.
    """

    @property
    def dims(self):
        return tuple(self.keys())

    @property
    def ndim(self):
        return len(self)

    def __repr__(self):
        s = "Coordinates:\n"
        for dim, coord in self.items():
            s += f"  * {dim}: "
            s += repr(coord) + "\n"
        return s

    def get_query(self, item):
        query = {dim: slice(None) for dim in self.dims}
        if isinstance(item, dict):
            query.update(item)
        elif isinstance(item, tuple):
            for k in range(len(item)):
                query[self.dims[k]] = item[k]
        else:
            query[self.dims[0]] = item
        return query

    def to_index(self, item):
        query = self.get_query(item)
        return {dim: self[dim].to_index(query[dim]) for dim in query}


class Coordinate:
    """
    Array-like object used to represent piecewise evenly spaced coordinates using the
    CF convention.

    The coordinate ticks are describes by the mean of tie points that are interpolated
    when intermediate values are required. Coordinate objects provides label based
    selections methods.

    Parameters
    ----------
    tie_indices : sequence of integers
        The indices of the tie points. Must include index 0 and be strictly increasing.
    tie_values : sequence of float or datetime64
        The values of the tie points. Must be strictly increasing to enable label-based
        selection. The len of `tie_indices` and `tie_values` sizes must match.
    """

    def __init__(self, tie_indices, tie_values):
        self.tie_indices = np.asarray(tie_indices)
        self.tie_values = np.asarray(tie_values)
        self.kind = "linear"

    def __bool__(self):
        if len(self.tie_indices) == 0 or len(self.tie_values) == 0:
            return False
        else:
            return True

    def __len__(self):
        if self:
            return self.tie_indices[-1] - self.tie_indices[0] + 1
        else:
            return 0

    def __repr__(self):
        if len(self) == 0:
            return "empty coordinate"
        elif len(self) == 1:
            return f"one point at {self.tie_values[0]}"
        else:
            return (
                f"{len(self.tie_indices)} tie points from {self.tie_values[0]} to "
                f"{self.tie_values[-1]}"
            )

    def __eq__(self, other):
        return np.array_equal(self.tie_indices, other.tie_indices) and np.array_equal(
            self.tie_values, other.tie_values
        )

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.slice_index(item)
        else:
            return self.get_value(item)

    def __array__(self):
        return self.values

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        raise NotImplementedError()

    def __array_function__(self, func, types, args, kwargs):
        raise NotImplementedError()

    @property
    def dtype(self):
        return self.tie_values.dtype

    @property
    def ndim(self):
        return 1

    @property
    def shape(self):
        return (len(self),)

    @property
    def indices(self):
        return np.arange(self.tie_indices[-1] + 1)

    @property
    def values(self):
        return self.get_value(self.indices)

    def format_index(self, idx, bounds="raise"):
        idx = np.asarray(idx)
        if not np.issubdtype(idx.dtype, np.integer):
            raise IndexError("only integer are valid index")
        idx = idx + (idx < 0) * len(self)
        if bounds == "raise":
            if np.any(idx < 0) or np.any(idx >= len(self)):
                raise IndexError("index is out of bounds")
        elif bounds == "clip":
            idx = np.clip(idx, 0, len(self))
        return idx

    def format_index_slice(self, slc):
        start = slc.start
        stop = slc.stop
        step = slc.step
        if start is None:
            start = 0
        if stop is None:
            stop = len(self)
        if step is None:
            step = 1
        start = self.format_index(start, bounds="clip")
        stop = self.format_index(stop, bounds="clip")
        return slice(start, stop, step)

    def get_value(self, index):
        index = self.format_index(index)
        return linear_interpolate(index, self.tie_indices, self.tie_values)

    def get_index(self, value, method=None):
        if isinstance(value, str):
            value = np.datetime64(value)
        else:
            value = np.asarray(value)
        if method is None:
            index = linear_interpolate(value, self.tie_values, self.tie_indices)
            index = np.rint(index).astype("int")
            index_value = self.get_value(index)
            if np.issubdtype(self.dtype, np.datetime64):
                if not np.all(index_value == value):
                    raise KeyError("value not found in index")
                else:
                    return index
            else:
                if not np.allclose(index_value, value):
                    raise KeyError("value not found in index")
                else:
                    return index
        elif method == "nearest":
            index = linear_interpolate(value, self.tie_values, self.tie_indices)
            return np.rint(index).astype("int")
        elif method == "before":
            index = linear_interpolate(
                value, self.tie_values, self.tie_indices, left=np.nan
            )
            if np.any(np.isnan(index)):
                raise KeyError("value not found in index")
            else:
                return np.floor(index).astype("int")
        elif method == "after":
            index = linear_interpolate(
                value, self.tie_values, self.tie_indices, right=np.nan
            )
            if np.any(np.isnan(index)):
                raise KeyError("value not found in index")
            else:
                return np.ceil(index).astype("int")
        else:
            raise ValueError("valid methods are: 'nearest', 'before', 'after'")

    def get_index_slice(self, value_slice):
        if value_slice.start is None:
            start = None
        else:
            try:
                start = self.get_index(value_slice.start, method="after")
            except KeyError:
                start = len(self)
        if value_slice.stop is None:
            stop = None
        else:
            try:
                end = self.get_index(value_slice.stop, method="before")
                stop = end + 1
            except KeyError:
                stop = 0
        return slice(start, stop)

    def slice_index(self, index_slice):
        index_slice = self.format_index_slice(index_slice)
        start_index, stop_index, step_index = (
            index_slice.start,
            index_slice.stop,
            index_slice.step,
        )
        if stop_index - start_index <= 0:
            return Coordinate([], [])
        elif (stop_index - start_index) <= step_index:
            tie_indices = [0]
            tie_values = [self.get_value(start_index)]
            return Coordinate(tie_indices, tie_values)
        else:
            end_index = stop_index - 1
            start_value = self.get_value(start_index)
            end_value = self.get_value(end_index)
            mask = (start_index < self.tie_indices) & (self.tie_indices < end_index)
            tie_indices = np.insert(
                self.tie_indices[mask],
                (0, self.tie_indices[mask].size),
                (start_index, end_index),
            )
            tie_values = np.insert(
                self.tie_values[mask],
                (0, self.tie_values[mask].size),
                (start_value, end_value),
            )
            tie_indices -= tie_indices[0]
            coord = Coordinate(tie_indices, tie_values)
            if step_index != 1:
                coord = coord.decimate(step_index)
            return coord

    def decimate(self, q):
        tie_indices = (self.tie_indices // q) * q
        for k in range(1, len(tie_indices) - 1):
            if tie_indices[k] == tie_indices[k - 1]:
                tie_indices[k] += q
        tie_values = [self.get_value(idx) for idx in tie_indices]
        tie_indices //= q
        return self.__class__(tie_indices, tie_values)

    def to_index(self, item):
        if isinstance(item, slice):
            return self.get_index_slice(item)
        else:
            return self.get_index(item)

    def simplify(self, tolerance):
        tie_indices, tie_values = douglas_peucker(
            self.tie_indices, self.tie_values, tolerance
        )
        return self.__class__(tie_indices, tie_values)

    def get_discontinuities(self):
        (indices,) = np.nonzero(np.diff(self.tie_indices) == 1)
        return [
            {
                self.tie_indices[index]: self.tie_values[index],
                self.tie_indices[index + 1]: self.tie_values[index + 1],
            }
            for index in indices
        ]


class ScaleOffset:
    def __init__(self, scale, offset):
        self.scale = scale
        self.offset = offset

    def __eq__(self, other):
        try:
            return (self.scale - other.scale == 0) and (self.offset - other.offset == 0)
        except:
            return False

    @classmethod
    def floatize(cls, arr):
        arr = np.asarray(arr)
        if np.issubdtype(arr.dtype, np.datetime64):
            unit, count = np.datetime_data(arr.dtype)
            scale = np.timedelta64(count, unit)
            offset = np.min(arr) + (np.max(arr) - np.min(arr)) / 2
        else:
            scale = 1.0
            offset = 0.0
        transform = cls(scale, offset)
        transform.check_resolution(arr)
        return transform

    def direct(self, arr):
        arr = np.asarray(arr)
        self.check_resolution(arr)
        return (arr - self.offset) / self.scale

    def inverse(self, arr):
        arr = np.asarray(arr)
        if np.issubdtype(np.asarray(self.scale).dtype, np.timedelta64):
            arr = np.rint(arr)
        return self.scale * arr + self.offset

    def check_resolution(self, arr):
        arr = np.asarray(arr)
        nmax = 2 ** np.finfo("float").nmant
        if not np.all((arr - self.offset).astype("int") < nmax):
            warnings.warn(
                "float resolution is not sufficient to represent the full integer range"
            )


def linear_interpolate(x, xp, fp, left=None, right=None):
    if not is_strictly_increasing(xp):
        raise ValueError("xp must be strictly increasing")
    x_transform = ScaleOffset.floatize(xp)
    f_transform = ScaleOffset.floatize(fp)
    x = x_transform.direct(x)
    xp = x_transform.direct(xp)
    fp = f_transform.direct(fp)
    f = np.interp(x, xp, fp, left=left, right=right)
    f = f_transform.inverse(f)
    return f


def is_strictly_increasing(x):
    if np.issubdtype(x.dtype, np.datetime64):
        return np.all(np.diff(x) > np.timedelta64(0, "us"))
    else:
        return np.all(np.diff(x) > 0)


def douglas_peucker(x, y, epsilon):
    mask = np.ones(len(x), dtype=bool)
    stack = [(0, len(x))]
    while stack:
        start, stop = stack.pop()
        ysimple = linear_interpolate(
            x[start:stop],
            x[[start, stop - 1]],
            y[[start, stop - 1]],
        )
        d = np.abs(y[start:stop] - ysimple)
        index = np.argmax(d)
        dmax = d[index]
        index += start
        if dmax > epsilon:
            stack.append([start, index + 1])
            stack.append([index, stop])
        else:
            mask[start + 1 : stop - 1] = False
    return x[mask], y[mask]



import tempfile

import numpy as np
import xarray as xr

import xdas


class TestAll:
    def test_all(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            # generation
            shape = (6000, 1000)
            resolution = (np.timedelta64(10, "ms"), 5.0)
            starttimes = {
                "001": np.datetime64("2023-01-01T00:00:00"),
                "002": np.datetime64("2023-01-01T00:01:00"),
                "003": np.datetime64("2023-01-01T00:02:00"),
            }
            for name, starttime in starttimes.items():
                db = xdas.Database(
                    data=np.random.randn(*shape).astype("float32"),
                    coords={
                        "time": xdas.Coordinate(
                            tie_indices=[0, shape[0] - 1],
                            tie_values=[
                                starttime,
                                starttime + resolution[0] * (shape[0] - 1),
                            ],
                        ),
                        "distance": xdas.Coordinate(
                            tie_indices=[0, shape[1] - 1],
                            tie_values=[0.0, resolution[1] * (shape[1] - 1)],
                        ),
                    },
                )
                db.to_netcdf(f"{tmpdir}/{name}.nc")

            # testing
            _db = xdas.open_database(f"{tmpdir}/002.nc")
            _da = _db.to_xarray()
            db = _db.sel(
                time=slice("2023-01-01T00:01:20", None),
                distance=slice(1000, None),
            )
            db = db.isel(
                time=slice(None, 1000),
                distance=slice(None, 200),
            )
            da1 = db.to_xarray()
            da = _da.sel(
                time=slice("2023-01-01T00:01:20", None),
                distance=slice(1000, None),
            )
            da = da.isel(
                time=slice(None, 1000),
                distance=slice(None, 200),
            )
            da2 = da
            xr.testing.assert_equal(da1, da2)



import numpy as np
import xarray as xr

import xdas
import xdas.signal as xp


class TestSignal:
    def test_get_sample_spacing(self):
        s = (5.0 / 2) + 5.0 * np.arange(100)
        dt = np.timedelta64(8, "ms")
        t = np.datetime64(0, "s") + dt * np.arange(1000)
        da = xr.DataArray(np.ones((len(t), len(s))), {"time": t, "distance": s})
        assert xp.get_sample_spacing(da, "time") == 0.008
        assert xp.get_sample_spacing(da, "distance") == 5.0
        shape = (6000, 1000)
        resolution = (np.timedelta64(8, "ms"), 5.0)
        starttime = np.datetime64("2023-01-01T00:00:00")
        db = xdas.Database(
            data=np.random.randn(*shape).astype("float32"),
            coords={
                "time": xdas.Coordinate(
                    tie_indices=[0, shape[0] - 1],
                    tie_values=[starttime, starttime + resolution[0] * (shape[0] - 1)],
                ),
                "distance": xdas.Coordinate(
                    tie_indices=[0, shape[1] - 1],
                    tie_values=[0.0, resolution[1] * (shape[1] - 1)],
                ),
            },
        )
        assert xp.get_sample_spacing(db, "time") == 0.008
        assert xp.get_sample_spacing(db, "distance") == 5.0

    def test_integrate(self):
        n = 100
        d = 5.0
        s = (d / 2) + d * np.arange(n)
        da = xr.DataArray(np.ones(n), {"distance": s})
        da = xp.integrate(da, midpoints=True)
        assert np.allclose(da, da["distance"])

    def test_segment_mean_removal(self):
        n = 100
        d = 5.0
        s = (d / 2) + d * np.arange(n)
        limits = [0, 0.3 * n * d, n * d]
        s = np.linspace(0, 1000, n)
        data = np.zeros(n)
        da = xr.DataArray(data, {"distance": s})
        da.loc[{"distance": slice(limits[0], limits[1])}] = 1.0
        da.loc[{"distance": slice(limits[1], limits[2])}] = 2.0
        da = xp.segment_mean_removal(da, limits)
        assert np.allclose(da, 0)

    def test_sliding_window_removal(self):
        n = 100
        d = 5.0
        s = (d / 2) + d * np.arange(n)
        s = np.linspace(0, 1000, n)
        data = np.ones(n)
        da = xr.DataArray(data, {"distance": s})
        da = xp.sliding_mean_removal(da, 0.1 * n * d)
        assert np.allclose(da, 0)



import numpy as np

import xdas


class TestCore:
    def test_core(self):
        shape = (300, 100)
        db = xdas.Database(
            data=np.random.randn(*shape),
            coords={
                "time": xdas.Coordinate([0, shape[0] - 1], [0, 3.0 - 1 / 100]),
                "distance": xdas.Coordinate([0, shape[1] - 1], [0, 990.0]),
            },
        )
        dbs = [db[100 * k : 100 * (k + 1)] for k in range(3)]
        _db = xdas.concatenate(dbs)
        assert np.array_equal(_db.data, db.data)
        assert _db["time"] == _db["time"]
        shape = (300, 100)
        db = xdas.Database(
            data=np.random.randn(*shape),
            coords={
                "time": xdas.Coordinate(
                    [0, shape[0] - 1],
                    [np.datetime64(0, "ms"), np.datetime64(2990, "ms")],
                ),
                "distance": xdas.Coordinate([0, shape[1] - 1], [0, 990.0]),
            },
        )
        dbs = [db[100 * k : 100 * (k + 1)] for k in range(3)]
        _db = xdas.concatenate(dbs)
        assert np.array_equal(_db.data, db.data)
        assert _db["time"] == _db["time"]
        dbs = [db[:, 20 * k : 20 * (k + 1)] for k in range(5)]
        _db = xdas.concatenate(dbs, "distance")
        assert np.array_equal(_db.data, db.data)
        assert _db["distance"] == _db["distance"]



import dask.array as da
import numpy as np
import scipy.signal as sp

from xdas.core import Coordinate, Database
from xdas.processing import LFilter, SignalProcessingChain, SOSFilter


class TestProcessing:
    def test_processing(self):
        fs = 125
        ks = 1 / 10
        data = da.random.normal(size=(1000, 100))
        time = Coordinate([0, data.shape[0] - 1], [0.0, (data.shape[0] - 1) / fs])
        distance = Coordinate([0, data.shape[1] - 1], [0.0, (data.shape[1] - 1) / ks])
        xarr = Database(data, {"time": time, "distance": distance})
        b, a = sp.iirfilter(4, 0.5, btype="lowpass")
        lfilter = LFilter(b, a, "time")
        result_direct = lfilter(xarr)
        chunk_size = 100
        lfilter.reset()
        result_chunks = xarr.copy()
        for k in range(xarr.shape[0] // chunk_size):
            query = {"time": slice(k * chunk_size, (k + 1) * chunk_size)}
            result_chunks[query] = lfilter(xarr[query]).data
        lfilter.reset()
        chain = SignalProcessingChain([lfilter])
        out = chain.process(xarr, "time", chunk_size, parallel=False)
        lfilter.reset()
        assert chain.filters[0].zi == None
        out_parallel = chain.process(xarr, "time", chunk_size, parallel=True)
        assert np.allclose(result_chunks.data, result_direct.data)
        assert np.allclose(out.data, result_direct.data)
        assert np.allclose(out_parallel.data, result_direct.data)

        sos = sp.iirfilter(4, 0.5, btype="lowpass", output="sos")
        sosfilter = SOSFilter(sos, "time")
        result_direct = sosfilter(xarr)
        chunk_size = 100
        sosfilter.reset()
        result_chunks = xarr.copy()
        for k in range(xarr.shape[0] // chunk_size):
            query = {"time": slice(k * chunk_size, (k + 1) * chunk_size)}
            result_chunks[query] = sosfilter(xarr[query]).data
        sosfilter.reset()
        chain = SignalProcessingChain([sosfilter])
        out = chain.process(xarr, "time", chunk_size, parallel=False)
        sosfilter.reset()
        assert chain.filters[0].zi == None
        out_parallel = chain.process(xarr, "time", chunk_size, parallel=True)
        assert np.allclose(result_chunks.data, result_direct.data)
        assert np.allclose(out.data, result_direct.data)
        assert np.allclose(out_parallel.data, result_direct.data)



import numpy as np
import scipy.signal as sp
import xarray as xr


def get_sample_spacing(da, dim):
    """
    Returns the sample spacing along a given dimension.

    Parameters
    ----------
    da : DataArray or Database
        The data from which extract the sample spacing.
    dim : str
        The dimension along which get the sample spacing.

    Returns
    -------
    float
        The sample spacing.
    """
    d = (da[dim][-1] - da[dim][0]) / (len(da[dim]) - 1)
    d = np.asarray(d)
    if np.issubdtype(d.dtype, np.timedelta64):
        d = d / np.timedelta64(1, "s")
    d = d.item()
    return d


def integrate(da, midpoints=False, dim="distance"):
    """
    Integrate along a given dimension.

    Parameters
    ----------
    da : DataArray
        The data to integrate.
    midpoints : bool, optional
        Whether to move the coordinates by half a step, by default False.
    dim : str, optional
        The dimension along which to integrate, by default "distance".

    Returns
    -------
    DataArray
        The integrated data.
    """
    d = get_sample_spacing(da, dim)
    out = da.cumsum(dim) * d
    if midpoints:
        s = da[dim].values
        s += d / 2
        out[dim] = s
    return out


def segment_mean_removal(da, limits, window="hann", dim="distance"):
    """
    Piecewise mean removal.

    Parameters
    ----------
    da : DataArray
        The data that segment mean should be removed.
    limits : list of float
        The segments limits.
    window : str, optional
        The tapering windows to apply at each window, by default "hann".
    dim : str, optional
        The axis along which remove the segment means, by default "distance".

    Returns
    -------
    DataArray
        The data with segment means removed.
    """
    out = da.copy()
    for sstart, send in zip(limits[:-1], limits[1:]):
        key = dict(distance=slice(sstart, np.nextafter(send, -np.inf)))
        subset = out.loc[key]
        win = xr.DataArray(
            sp.get_window(window, subset.sizes[dim]),
            {dim: subset[dim]},
        )
        ref = (subset * win).sum(dim) / win.sum(dim)
        out.loc[key] -= ref
    return out


def sliding_mean_removal(da, wlen, window="hann", pad_mode="reflect", dim="distance"):
    """
    Sliding mean removal.

    Parameters
    ----------
    da : DataArray
        The data that sliding mean should be removed.
    wlen : float
        Length of the sliding mean.
    window : str, optional
        Tapering window used, by default "hann"
    pad_mode : str, optional
        Padding mode used, by default "reflect"
    dim : str, optional
        The dimension along which remove the sliding mean, by default "distance"

    Returns
    -------
    DataArray
        The data with sliding mean removed.
    """
    d = get_sample_spacing(da, dim)
    n = round(wlen / d)
    if n % 2 == 0:
        n += 1
    win = sp.get_window(window, n)
    win /= np.sum(win)
    shape = tuple(-1 if d == dim else 1 for d in da.dims)
    win = np.reshape(win, shape)
    data = da.data
    pad_width = tuple((n // 2, n // 2) if d == dim else (0, 0) for d in da.dims)
    mean = sp.fftconvolve(np.pad(data, pad_width, mode=pad_mode), win, mode="valid")
    data = data - mean
    return da.copy(data=data)
